#!/usr/bin/env ruby

# I use  because I want to use the RUDL.so that is made by extconf.rb and the makefile.
# You normally use just "require 'RUDL'".
require 'RUDL'

# Namespaces are boring, let's get rid of them.
include RUDL
include Constant

# SWSURFACE, because (it looks like) a hardwaresurface sometimes isn't
# doublebuffered, and that meansthat we'll be drawing directly to the screen, 
# which is ugly. So I always ask for a software surface.
$display=DisplaySurface.new([320, 200], SWSURFACE|FULLSCREEN)

# contained_images parses the surface made from "crapola.bmp" in the following way:
# 1. get color on 0,0 - call it bordercolor
# 2. Start going right, until another bordercolor is found.
# 3. Go down, until yet another bordercolor is found.
# 4. Between these positions, there's an image. Make a new surface with it,
#    and store it in the 2-dimensional result array.
# 5. Repeat 2 until you run out of pixels.
# 6. Find the next "line" of images and go back to 2.
# 7. Return the two-dimensional array of found images
# This might be easier to see if you look at crapola.bmp.
$images=Surface.load_new('crapola.bmp').contained_images

$score=0

# Give the monitor some time to switch modes.
Timer.delay(1000)

# Converts all images to match the display's color type, 
# and set black to be the transparent color.
def prepare_images
	$images.each do |line|
		line.each do |image|
			image.convert
			image.set_colorkey([0,0,0], RLEACCEL)
		end
	end
end

# A """"sprite engine""""
class Sprite

	# All sprites that are active
	@@spritelist=[]

	# For collision detection
	@@goodbulletlist=[]
	@@badthinglist=[]
	@@goodthinglist=[]

	def Sprite.update
		bulletrect=nil
		
		@@spritelist.each do |sprite|
			sprite.update
		end

		Rect.collide_lists(@@goodbulletlist, @@badthinglist) do |bullet, thing|
			bullet.stop
			thing.stop
			25.times do Explosion.new([thing.rect.x, thing.rect.y]) end
			$score=$score+10
		end

		Rect.collide_lists(@@badthinglist, @@goodthinglist) do |enemy, friend|
			enemy.stop
			friend.stop
			25.times do Explosion.new([enemy.rect.x, enemy.rect.y]) end
		end
	end
	
	def Sprite.draw
		@@spritelist.each do |sprite|
			sprite.draw
		end
	end

	def Sprite.reset
		@@spritelist=[]
		@@goodbulletlist=[]
		@@badthinglist=[]
		@@goodthinglist=[]
	end

	attr_accessor :rect, :image

	def initialize(position, image)
		@image=image
		@rect=Rect.new(position[0], position[1], @image.w, @image.h)
		start
	end

	def update
	end

	def start
		@@spritelist.push(self)
	end

	def stop
		@@spritelist.delete(self)
	end

	def draw
		$display.blit(@image, [@rect.x, @rect.y])
	end
end

class Explosion < Sprite
	def initialize(position)
		super(position, $images[0][rand(3)+4])
		@direction=[(rand(99)-49)/10.0, (rand(99)-49)/10.0]
		@timetolive=rand(10)
	end

	def update
		@timetolive-=1
		if @timetolive<0 then
			stop
		end
		@rect.x+=@direction[0]
		@rect.y+=@direction[1]
		super
	end
end

class Bullet < Sprite
	def initialize(position)
		bulletposition=[position[0]+3, position[1]-4]
		super(bulletposition, $images[0][1])
	end

	def update
		@rect.y-=2
		if @rect.y<-30 then stop end
		super
	end

	def start
		if @@goodbulletlist.size<9 then
			@@goodbulletlist.push(self)
			super
		end
	end

	def stop
		@@goodbulletlist.delete(self)
		super
	end
end

class Rock < Sprite
	def initialize
		super([rand(300), -30], $images[0][2])
		@direction=[(rand(99)-49)/20.0, rand(2)+1]
	end

	def update
		@rect.x+=@direction[0]
		@rect.y+=@direction[1]
		stop if @rect.y>200
		@direction[0]=@direction[0].abs if @rect.x<0
		@direction[0]=-(@direction[0].abs) if @rect.x>290
		super
	end

	def start
		if @@badthinglist.size<50
			super
			@@badthinglist.push(self)
		end
	end

	def stop
		super
		@@badthinglist.delete(self)
	end
end

class Enemy < Sprite
	def initialize
		if rand(2)==1
			super([320, rand(100)], $images[0][3])
			@direction=-rand(5)
		else
			super([-30, rand(100)], $images[0][3])
			@direction=rand(5)
		end
		@center=rand(320)
	end

	def update
		@rect.x+=@direction
		@direction+=1 if @rect.x<@center-50
		@direction-=1 if @rect.x>@center+50
		@direction=-3 if @direction<-3
		@direction=3 if @direction>3
		@center=rand(320) if rand(40)==0
		if rand(190)==0
			Bomb.new(@rect)
		end
		super
	end

	def start
		if @@badthinglist.size<50
			super
			@@badthinglist.push(self)
		end
	end

	def stop
		@@badthinglist.delete(self)
		DestroyedEnemy.new(@rect, @direction)
		super
	end
end

class DestroyedEnemy < Sprite
	def initialize(posrect, direction)
		super([0,0], $images[0][3])
		@rect=posrect
		@direction=direction
	end

	def update
		@rect.x+=@direction
		@rect.y+=1
		stop if @rect.y>200
		Explosion.new([@rect.x, @rect.y])
	end
end

class Bomb < Sprite
	def initialize(posrect)
		super([0,0], $images[0][4])
		@rect=posrect.clone
	end

	def update
		@rect.y+=1
		stop if @rect.y>200
	end

	def start
		if @@badthinglist.size<50
			super
			@@badthinglist.push(self)
		end
	end

	def stop
		super
		@@badthinglist.delete(self)
	end
end

class Ship < Sprite
	attr_accessor :dead

	def initialize
		super([160, 160], $images[0][0])
		@endlessbulletstreamstopper=false
		@dead=false
	end

	def update
		# This is a bit ugly, the keystate should be grabbed once every frame
		# and be stored somewhere so that everything can access it.
		keystate=Key.pressed?

		if keystate[K_LEFT] && @rect.x>0
			@rect.x-=2
		end

		if keystate[K_RIGHT] && @rect.x<300
			@rect.x+=2
		end
		
		if keystate[K_LCTRL] 
			if !@endlessbulletstreamstopper
				Bullet.new([@rect.x, @rect.y])
				@endlessbulletstreamstopper=true
			end
		else
			@endlessbulletstreamstopper=false
		end

		super
	end

	def start
		if @@goodthinglist.size<50
			super
			@@goodthinglist.push(self)
		end
	end

	def stop
		super
		@@goodthinglist.delete(self)
		@dead=true
	end
end

def play
	Sprite.reset  # Empty spritelists
	ship=Ship.new

	lives=5
	$score=0
	deadtimer=-1  # Statevariable: if bigger than zero, do "GET READY" bit

	moment=Timer.ticks

	running=1.0
	frameskip=1

	while lives>0

		while event=EventQueue.poll
			case event
				when QuitEvent
					exit
				when KeyDownEvent
					case event.key
						when K_ESCAPE
							exit
						when K_RETURN
							if event.mod & KMOD_ALT>0
								$display=DisplaySurface.new([320, 200], DOUBLEBUF|($display.flags^FULLSCREEN))
								prepare_images
							end
					end
			end
		end


		# I'm not sure if this is a very optimal timing scheme:
		passed=(Timer.ticks-moment)/10.0
		if passed>5 then passed=5 end  # to avoid gigantic leaps forward in time.
		if passed<1 then passed=1 end
		moment=Timer.ticks

		running=(running*3+passed)/4
		frameskip=running.ceil if running>frameskip+0.5
		frameskip=running.floor if running<frameskip-0.5


		# Start a new frame
		$display.fill([0, 0, 0])

		# Update as many times as frames have passed.
		# Actually, frameskip is not counted in frames but in looks-good-enough units.
		frameskip.times do
			Sprite.update
			if deadtimer>0
				$display.blit($images[3][0], [(320-$images[3][0].w)/2, (200-$images[3][0].h)/2])
				deadtimer=deadtimer-1
				if deadtimer==0
					ship=Ship.new
				end
			else
				if ship.dead
					deadtimer=200
					lives-=1
					return if lives<=0
				end
			end

			Rock.new if rand(20)==0
			Enemy.new if rand(50)==0
		end

		# blit lives
		$display.blit($images[5][0], [0,2])
		(1..lives).each do |nr|
			$display.blit($images[0][0], [30+16*nr, 0])
		end

		# blit score
		$display.blit($images[1][0], [0,20])
		scoretmp=$score
		digitx=100
		while scoretmp>0
			$display.blit($images[2][scoretmp%10], [digitx, 20])
			digitx=digitx-10
			scoretmp=scoretmp/10
		end

		# blit all sprites in the sprite "engine"
		Sprite.draw

		# show it
		$display.flip
	end
end


def gameover
	start=false
	gameoverimg=$images[4][0]
	while !start
		while event=EventQueue.poll
			case event
				when QuitEvent
					exit
				when KeyDownEvent
					case event.key
						when K_ESCAPE
							exit
						when K_LCTRL
							start=true
						when K_RETURN
							if event.mod & KMOD_ALT>0
								$display=DisplaySurface.new([320, 200], SWSURFACE|($display.flags^FULLSCREEN))
								prepare_images
							end
					end
			end
		end
		$display.fill([0, 0, 0])
		$display.blit($images[4][0], [(320-gameoverimg.w)/2, (200-gameoverimg.h)/2])
		$display.flip
	end
end

prepare_images

while true
	gameover
	play
end
